import os

import pytest


@pytest.mark.skipif(
    not os.getenv("FAST_BITRIX24_TEST_WEBHOOK"),
    reason="Нет аккаунта, на котором можно проверить",
)
class TestsWithLiveServer:
    class TestBasic:
        def test_simple_add_lead(self, get_test):
            b = get_test
            lead_no = b.call(
                "crm.lead.add", {"fields": {"NAME": "Bob", "COMMENTS": "x" * 100}}
            )
            b.get_by_ID("crm.lead.delete", [lead_no])

        def test_simple_get_all(self, create_100_leads):
            b = create_100_leads

            result = b.get_all("crm.lead.list")
            assert isinstance(result, list)
            assert len(result) >= 100

        def test_get_all_single_page(self, get_test):
            b = get_test

            result = b.get_all("crm.lead.fields")
            assert isinstance(result, dict)

        def test_get_all_params(self, create_100_leads):
            b = create_100_leads

            fields = ["ID", "NAME"]

            leads = b.get_all("crm.lead.list", {"select": fields})

            assert len(fields) == len(leads[0])

        def test_issue_96(self, get_test):
            from datetime import datetime

            b = get_test
            b.call(
                "telephony.externalcall.register",
                {
                    # 'USER_PHONE_INNER': 'OLD_LINE',
                    "USER_ID": 1,
                    "PHONE_NUMBER": "+79163345641",
                    "CALL_START_DATE": f"{datetime.now()}",
                    "CRM_CREATE": 0,
                    "CRM_ENTITY_TYPE": "LEAD",
                    "CRM_ENTITY_ID": 43707,
                    "SHOW": 0,
                    "TYPE": 1,
                },
            )

        def test_call_batch(self, create_100_leads):
            b = create_100_leads

            with pytest.raises(ValueError):
                b.call_batch({})

            result = b.call_batch({"halt": 0, "cmd": {1: "crm.lead.list"}})

            assert len(result) == 1
            assert len(result["1"]) == 50

        def test_get_by_ID_results(self, create_100_leads):
            b = create_100_leads

            leads = b.get_all("crm.lead.list")
            lead_IDs = [lead["ID"] for lead in leads[:10]]

            leads = b.get_by_ID("crm.lead.get", lead_IDs)
            assert isinstance(leads, dict)

        def test_call(self, get_test):
            b = get_test

            delete_IDs = b.call(
                "crm.lead.add",
                [
                    {"fields": {"NAME": "Bob"}},
                    {"fields": {"NAME": "Jake"}},
                ],
            )
            b.call("crm.lead.delete", [{"ID": ID} for ID in delete_IDs])

            b.call("crm.lead.delete", [])

        def test_call_single_param(self, get_test):
            b = get_test
            delete_ID = b.call("crm.lead.add", {"fields": {"NAME": "Bob"}})
            b.call("crm.lead.delete", {"ID": delete_ID})

        def test_issue_129(self, create_a_lead):
            b, lead = create_a_lead

            # итератор и прогресс-бар - ошибка
            with pytest.raises(TypeError):
                b.get_by_ID("crm.lead.get", iter([lead]))

            with pytest.raises(TypeError):
                b.call("crm.lead.get", iter([{"ID": lead}]))

            # итератор и без прогресс бара - нет ошибки
            b.verbose = False
            b.get_by_ID("crm.lead.get", iter([lead]))
            b.call("crm.lead.get", iter([{"ID": lead}]))

            # sequence - нет ошибки
            b.verbose = True
            b.get_by_ID("crm.lead.get", [lead])

            result = b.call("crm.lead.get", [{"ID": lead}])
            assert isinstance(result, tuple)

            result = b.call("crm.lead.get", {"ID": lead})
            assert isinstance(result, dict)

            # пустой итератор и без прогресс бара - нет ошибки
            b.verbose = False
            b.get_by_ID("crm.lead.get", iter([]))
            b.call("crm.lead.get", iter([]))

        def test_issue_132(self, create_100_tasks):
            b = create_100_tasks

            result = b.get_all("tasks.task.list")
            assert result

            with pytest.raises(ValueError):
                result = b.list_and_get("tasks.task", "taskId")

        def test_case(self, get_test):
            b = get_test

            with pytest.raises(
                RuntimeError, match="Could not find value for parameter"
            ):
                b.call("disk.file.get", [{"ID": 1}])

            with pytest.raises(RuntimeError, match="Could not find entity with id"):
                b.call("disk.file.get", [{"id": 1}])

        def test_long_task_description(self, get_test):
            b = get_test
            lead_no = b.call(
                "crm.lead.add", {"fields": {"NAME": "Bob", "COMMENTS": "x" * 10000}}
            )
            b.get_by_ID("crm.lead.delete", [lead_no])

    class TestParamsEncoding:
        def test_mobile_phone(self, get_test):
            b = get_test
            lead_no = b.call(
                "crm.lead.add",
                {
                    "fields": {
                        "NAME": "Bob",
                        "PHONE": [{"VALUE": "+7123456789", "VALUE_TYPE": "MOBILE"}],
                    }
                },
            )
            lead = b.get_by_ID("crm.lead.get", [lead_no])[str(lead_no)]

            try:
                assert lead["PHONE"][0]["VALUE_TYPE"] == "MOBILE"
            finally:
                b.get_by_ID("crm.lead.delete", [lead_no])

        def test_filter_not_equal(self, create_100_leads):
            b = create_100_leads

            result = b.get_all("crm.lead.list", {"FILTER": {"!STATUS_ID": "NEW"}})
            assert not result

            result = b.get_all("crm.lead.list", {"FILTER": {"!STATUS_ID": "CLOSED"}})
            assert result

            result = b.get_all("crm.lead.list", {"FILTER": {"<>STATUS_ID": "NEW"}})
            assert result

            result = b.get_all("crm.lead.list", {"FILTER": {"<>STATUS_ID": "CLOSED"}})
            assert result

        def test_product_rows(self, create_a_lead):
            b, lead_no = create_a_lead

            product_rows = [
                {
                    "PRODUCT_NAME": "ssssdsd",
                    "PRICE": 5555,
                    "QUANTITY": 2,
                    "CURRENCY": "USD",
                },
                {"PRODUCT_ID": 2809, "PRICE": 100, "QUANTITY": 2},
            ]

            b.call("crm.lead.productrows.set", {"ID": lead_no, "rows": product_rows})

            result_rows = b.call("crm.lead.productrows.get", {"ID": lead_no})

            assert len(product_rows) == len(result_rows)


class TestErrors:
    def test_get_all(self, bx_dummy):
        b = bx_dummy

        with pytest.raises(Exception):
            b.get_all("")

        with pytest.raises(Exception):
            b.get_all(123)

        with pytest.raises(Exception):
            b.get_all("some_method", {"select": None})

        with pytest.raises(Exception):
            b.get_all("some_method", {"filter": 3})

        with pytest.raises(Exception):
            b.get_all("task.elapseditem.getlist", {"filter": 3})

    def test_get_by_ID(self, bx_dummy):
        b = bx_dummy

        with pytest.raises(Exception):
            b.get_by_ID("_", 123)

        with pytest.raises(Exception):
            b.get_by_ID("_", [["a"]])

    def test_call(self, bx_dummy, monkeypatch):
        b = bx_dummy

        async def stub(*args, **kwargs):
            return {"result": {"result": {"ok"}}}

        monkeypatch.setattr(b.srh, "request_attempt", stub)
        assert b.srh.request_attempt is stub

        b.call("_", raw=True)

        with pytest.raises(Exception):
            b.call("_", {})

        with pytest.raises(Exception):
            b.call("_", {"select": ["*"]})

        b.call("_", [1, {"a": 2}], raw=True)



import asyncio
import math
import time

import pytest

from fast_bitrix24.throttle import LeakyBucketThrottler, SlidingWindowThrottler


# Test the acquire method of the LeakyBucketThrottler class
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "pool_size, requests_per_second, requests_made, sleep_time, test_id",
    [
        (5, 1.0, 3, 0, "acquire_happy_no_wait"),
        (5, 1.0, 8, 1, "acquire_happy_path_wait"),
    ],
)
async def test_leaky_bucket(
    pool_size, requests_per_second, requests_made, sleep_time, test_id, monkeypatch
):
    # Set up mocks
    start_time = time.monotonic()

    def fake_time():
        return start_time

    monkeypatch.setattr(time, "monotonic", fake_time)

    sleep_log = []

    async def fake_sleep(duration):
        sleep_log.append(duration)

    monkeypatch.setattr(asyncio, "sleep", fake_sleep)

    # Arrange
    throttler = LeakyBucketThrottler(pool_size, requests_per_second)
    for _ in range(requests_made):
        throttler.add_request_record()

    # Act
    async with throttler.acquire() as _:
        pass

    # Assert
    assert sum(sleep_log) == sleep_time


@pytest.mark.parametrize(
    "max_request_running_time, measurement_period, requests, measurements",
    [
        [2, 10, [], [[0, 0]]],
        [2, 10, [], [[5, 0]]],
        [2, 10, [], [[15, 0]]],
        [2, 10, [[0, 1], [2, 1]], [[5, 5]]],
        [2, 10, [[0, 1], [2, 1]], [[7, 3]]],
        [2, 10, [[0, 1], [2, 1], [3, 1]], [[5, 7]]],
        [
            2,
            10,
            [[0, 1], [2, 1], [3, 1], [10, 0.9]],
            [[0, 0], [1, 0], [2.1, 7.9], [10.1, 1.9]],
        ],
    ],
)
def test_needed_sleep_time(
    max_request_running_time,
    measurement_period,
    requests,
    measurements,
    monkeypatch,
):
    throttler = SlidingWindowThrottler(max_request_running_time, measurement_period)

    while requests or measurements:
        if (requests and measurements and requests[0][0] < measurements[0][0]) or (
            requests and not measurements
        ):
            when, duration = requests.pop(0)
            monkeypatch.setattr("time.monotonic", lambda: when)
            throttler.add_request_record(duration)
        else:
            call_point, expected = measurements.pop(0)
            monkeypatch.setattr("time.monotonic", lambda: call_point)
            print("Request record:", throttler._request_history)
            print("Time", call_point)
            assert math.isclose(throttler._calculate_needed_sleep_time(), expected)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "max_request_running_time, measurement_period, request_durations, expected_sleep_time, test_id",
    [
        # Happy path tests
        (10, 20, [1, 2, 3], 0, "happy-1"),
        (10, 20, [5, 5.1, 5], 9.9, "happy-2"),
        # Edge cases
        (10, 20, [10], 10, "edge-1"),
        (10, 20, [10, 0.1], 9.9, "edge-2"),
    ],
)
async def test_sliding_window(
    max_request_running_time,
    measurement_period,
    request_durations,
    expected_sleep_time,
    test_id,
    monkeypatch,
):
    # Set up mocks
    start_time = time.monotonic()

    def fake_time():
        return start_time

    monkeypatch.setattr(time, "monotonic", fake_time)

    sleep_log = []

    async def fake_sleep(duration):
        sleep_log.append(duration)

    monkeypatch.setattr(asyncio, "sleep", fake_sleep)

    # Arrange
    throttler = SlidingWindowThrottler(max_request_running_time, measurement_period)

    # Act
    for duration in request_durations:
        async with throttler.acquire():
            pass
        throttler.add_request_record(duration)
        start_time += duration
        await asyncio.sleep(duration)

    # Assert
    assert math.isclose(
        throttler._calculate_needed_sleep_time(), expected_sleep_time
    ), f"Test failed for {test_id}"



"""Высокоуровневый API для доступа к Битрикс24"""

import asyncio
import functools as ft
from contextlib import contextmanager
from inspect import iscoroutinefunction
from typing import Iterable, Union

import aiohttp
import icontract
from beartype import beartype

from . import correct_asyncio
from .logger import log, logger
from .server_response import ServerResponseParser
from .srh import ServerRequestHandler
from .user_request import (
    CallUserRequest,
    GetAllUserRequest,
    GetByIDUserRequest,
    ListAndGetUserRequest,
    RawCallUserRequest,
)


class BitrixAsync:
    """Клиент для асинхронных запросов к Битрикс24."""

    @beartype
    def __init__(
        self,
        webhook: str,
        token_func=None,
        verbose: bool = True,
        respect_velocity_policy: bool = True,
        request_pool_size: int = 50,
        requests_per_second: float = 2.0,
        batch_size: int = 50,
        operating_time_limit: int = 480,
        client: aiohttp.ClientSession = None,
        ssl: bool = True,
    ):
        """
        Создает объект для запросов к Битрикс24.

        Параметры:
        - `webhook: str` - URL вебхука, полученного от сервера Битрикс
        - `token_func: Awaitable = None` - асинхронная функция для получения
        и обновления токена в случае работы с приложениями и OAuth-авторизацией
        - `verbose: bool = True` - показывать ли прогрессбар при выполнении
        запроса
        - `respect_velocity_policy: bool = True` - соблюдать ли политику
        Битрикса о скорости запросов
        - `request_pool_size: int = 50` - размер пула запросов, который
        можно отправить на сервер без ожидания
        - `requests_per_second: float = 2.0` - максимальная скорость запросов,
        которая будет использоваться после переполнения пула
        - `batch_size: int = 50` - максимальное количество запросов, которые
        будут отправляться на сервер в одном батче
        - `operating_time_limit: int = 480` - максимальное допустимое время отработки
        запросов к одному методу REST API в секундах, допустимое за 10 минут,
        после которого запросы будут замедляться
        - `ssl: bool = True` - использовать ли проверку SSL-сертификата
        при HTTP-соединениях с сервером Битрикс.
        - `client: aiohttp.ClientSession = None` - использовать для HTTP-вызовов
        объект aiohttp.ClientSession, инициализированнный и настроенный
        пользователем. Ожидаеется, что пользователь сам откроет и закроет сессию.
        """

        if token_func is not None and not iscoroutinefunction(token_func):
            raise ValueError("`token_func` must be an async function.")

        self.srh = ServerRequestHandler(
            webhook=webhook,
            token_func=token_func,
            respect_velocity_policy=respect_velocity_policy,
            request_pool_size=request_pool_size,
            requests_per_second=requests_per_second,
            operating_time_limit=operating_time_limit,
            ssl=ssl,
            client=client,
        )
        self.verbose = verbose
        self.batch_size = batch_size

    @log
    async def get_all(self, method: str, params: dict = None) -> Union[list, dict]:
        """
        Получить полный список сущностей по запросу `method`.

        Под капотом использует параллельные запросы и автоматическое построение
        батчей, чтобы ускорить получение данных. Также самостоятельно
        обратывает постраничные ответы сервера, чтобы вернуть полный список.

        Параметры:
        - `method` - метод REST API для запроса к серверу
        - `params` - параметры для передачи методу. Используется
            именно тот формат, который указан в документации к REST API
            Битрикс24. `get_all()` не поддерживает параметры
            `start`, `limit` и `order`.

        Возвращает полный список сущностей, имеющихся на сервере,
        согласно заданным методу и параметрам.
        """

        return await self.srh.run_async(GetAllUserRequest(self, method, params).run())

    @log
    async def get_by_ID(
        self,
        method: str,
        ID_list: Iterable,
        ID_field_name: str = "ID",
        params: dict = None,
    ) -> dict:
        """
        Получить список сущностей по запросу `method` и списку ID.

        Используется для случаев, когда нужны не все сущности,
        имеющиеся в базе, а конкретный список поименованных ID.
        Например, все контакты, привязанные к сделкам.

        Параметры:
        - `method` - метод REST API для запроса к серверу
        - `ID_list` - список ID
        - `ID_field_name` - название поля, которое будет подаваться в запрос
        для каждого элемента ID_list
        - `params` - параметры для передачи методу. Используется именно тот
        формат, который указан в документации к REST API Битрикс24

        Возвращает словарь вида:
        ```
        {
            ID_1: <результат запроса 1>,
            ID_2: <результат запроса 2>,
            ...
        }
        ```

        Значением элемента словаря будет результат выполнения запроса
        относительно этого ID. Это может быть, например, список связанных
        сущностей или пустой список, если не найдено ни одной привязанной
        сущности.
        """

        return await self.srh.run_async(
            GetByIDUserRequest(self, method, params, ID_list, ID_field_name).run()
        )

    @log
    async def list_and_get(self, method_branch: str, ID_field_name="ID") -> dict:
        """
        Скачать список всех ID при помощи метода *.list,
        а затем все элементы при помощи метода *.get.

        Подобный подход показывает на порядок большую скорость
        получения данных, чем `get_all()`
        с параметром `'select': ['*', 'UF_*']`.

        Параметры:
        * `method_branch: str` - группа методов к использованию, например,
        `crm.lead` или `tasks.task`
        * `ID_field_name='ID'` - имя поля, в котором метод *.get принимает
        идентификаторы элементов (например, `'ID'` для метода `crm.lead.get`)

        Возвращает полное содержимое всех элементов в виде, используемом
        функцией `get_by_ID()` - словарь следующего вида:
        ```
        {
            ID_1: <словарь полей сущности с ID_1>,
            ID_2: <словарь полей сущности с ID_2>,
            ...
        }
        ```
        """

        return await ListAndGetUserRequest(
            self, method_branch, ID_field_name=ID_field_name
        ).run()

    @log
    async def call(
        self, method: str, items: Union[dict, Iterable] = None, *, raw=False
    ):
        """
        Вызвать метод REST API по списку элементов.

        Параметры:
        - `method` - метод REST API
        - `items` - список параметров вызываемого метода
            либо dict с параметрами для единичного вызова. Может быть `None`,
            если `raw=True`.
        - `raw` - если True, то items отправляются на сервер в виде json
            в первозданном виде, без обычных преобразований.
            По умолчанию False.

        Возвращает список ответов сервера для каждого из элементов `items`
        либо просто результат для единичного вызова.
        """

        request_cls = RawCallUserRequest if raw else CallUserRequest
        return await self.srh.run_async(request_cls(self, method, items).run())

    @log
    async def call_batch(self, params: dict) -> dict:
        """
        Вызвать метод `batch`.

        Параметры:
        - `params` - список параметров вызываемого метода

        Возвращает ответы сервера в формате словаря, где ключ - название
        команды, а значение - ответ сервера по этой команде.
        """

        response = ServerResponseParser(
            await self.srh.run_async(RawCallUserRequest(self, "batch", params).run())
        )

        response.raise_for_errors()

        return response.result["result"]

    @contextmanager
    @beartype
    @icontract.require(lambda max_concurrent_requests: max_concurrent_requests >= 1)
    def slow(self, max_concurrent_requests: int = 1):
        """Временно ограничивает количество одновременно выполняемых запросов
        к Битрикс24."""

        logger.info(
            "Slow mode enabled: {'max_concurrent_requests': %s}",
            max_concurrent_requests,
        )

        mcr_max_backup, self.srh.mcr_max = self.srh.mcr_max, max_concurrent_requests
        self.srh.mcr_cur_limit = min(self.srh.mcr_max, self.srh.mcr_cur_limit)

        yield True

        logger.info(
            "Slow mode disabled: {'max_concurrent_requests': %s}",
            mcr_max_backup,
        )

        self.srh.mcr_max = mcr_max_backup
        self.srh.mcr_cur_limit = min(self.srh.mcr_max, self.srh.mcr_cur_limit)


class Bitrix(BitrixAsync):
    """Клиент для неасинхронных запросов к серверу Битрикс24.

    Имплементируется путем обертки всех методов родителя в неасинхронные методы.
    """

    def sync_decorator(coroutine):
        ft.wraps(coroutine)

        def sync_wrapper(*args, **kwargs):
            try:
                loop = asyncio.get_event_loop()
            except RuntimeError:
                return asyncio.run(coroutine(*args, **kwargs))

            if loop.is_running():
                return loop.create_task(coroutine(*args, **kwargs))

            return loop.run_until_complete(coroutine(*args, **kwargs))

        return sync_wrapper

    for method in dir(BitrixAsync):
        if not method.startswith("__") and method != "slow":
            locals()[method] = sync_decorator(getattr(BitrixAsync, method))



import logging

import pytest
from beartype.typing import Dict, List, Union
from fast_bitrix24.logger import logger
from fast_bitrix24.server_response import (
    ErrorInServerResponseException,
    ServerResponseParser,
)
from fast_bitrix24.srh import ServerRequestHandler

logger.addHandler(logging.StreamHandler())


class MockSRH(ServerRequestHandler):
    def __init__(self, response: Union[Dict, List[Dict]]):
        self.response = response if isinstance(response, list) else [response]
        self.element_no = -1

        super().__init__("https://google.com/path", None, False, 50, 2, 480, None)

    async def single_request(self, *args, **kwargs):
        self.element_no += 1
        return self.response[self.element_no]


def test_single_success():
    from tests.real_responses.crm_lead_fields_success import response

    results = ServerResponseParser(response).extract_results()
    assert isinstance(results, dict)

    from tests.real_responses.crm_lead_list_one_page_success import response

    results = ServerResponseParser(response).extract_results()
    assert isinstance(results, list)


def test_batch_success(bx_dummy, recwarn):
    from tests.real_responses.crm_lead_list_several_pages_success import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_all("crm.lead.list")
    assert isinstance(results, list)
    print(w.message for w in recwarn.list)
    with pytest.raises(Exception):
        recwarn.pop(RuntimeWarning)


def test_batch_single_page_error(bx_dummy):
    from tests.real_responses.crm_get_batch_mix_success_and_errors import response

    bx_dummy.srh = MockSRH(response)
    with pytest.raises(ErrorInServerResponseException):
        bx_dummy.get_by_ID("crm.lead.get", [0, 1, 35943])


def test_call_single_success(bx_dummy):
    from tests.real_responses.call_single_success import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.call("crm.lead.get", {"ID": 35943})
    assert isinstance(results, dict)


def test_call_several_success(bx_dummy):
    from tests.real_responses.call_several_success import response

    bx_dummy.srh = MockSRH(response)

    ID_list = [161, 35943, 171]
    results = bx_dummy.call("crm.lead.get", [{"ID": ID} for ID in ID_list])

    assert isinstance(results, tuple)
    assert len(results) == 3
    assert isinstance(results[0], dict)

    # нужен какой-то другой тест для контроля порядока результатов
    # assert [result["ID"] for result in results] == ID_list, "Incorrect order of IDs"


def test_call_list_empty(bx_dummy):
    from tests.real_responses.batch_list_empty import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.call(
        "crm.lead.list", ({"filter": {"PHONE": "+0000877578564"}, "select": ["ID"]},)
    )
    assert isinstance(results, list)
    assert len(results) == 0

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.call(
        "crm.lead.list", {"filter": {"PHONE": "+0000877578564"}, "select": ["ID"]}
    )
    assert isinstance(results, list)
    assert len(results) == 0


def test_get_all_non_list_method(bx_dummy, recwarn):
    from tests.real_responses.user_fields import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_all("user.fields")
    assert isinstance(results, dict)
    print(w.message for w in recwarn.list)
    with pytest.raises(Exception):
        recwarn.pop(RuntimeWarning)


def test_batch_and_call_raw(bx_dummy):
    from tests.real_responses.call_batch import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.call_batch(
        {
            "halt": 0,
            "cmd": {
                "statuses": "crm.status.entity.items?entityId=STATUS",
                "sources": "crm.status.entity.items?entityId=SOURCE",
            },
        }
    )
    assert isinstance(results, dict)
    assert results.keys() == {"statuses", "sources"}

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.call(
        "batch",
        {
            "halt": 0,
            "cmd": {
                "statuses": "crm.status.entity.items?entityId=STATUS",
                "sources": "crm.status.entity.items?entityId=SOURCE",
            },
        },
        raw=True,
    )
    assert isinstance(results, dict)
    assert results.keys() == {"result", "time"}


def test_stagehistory(bx_dummy, recwarn):
    from tests.real_responses.stagehistory import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_all(
        "crm.stagehistory.list",
        {
            "entityTypeId": 2,
            "filter": {"STAGE_ID": "UC_KW9988"},
        },
    )

    assert isinstance(results, list)
    print(w.message for w in recwarn.list)
    with pytest.raises(Exception):
        recwarn.pop(RuntimeWarning)


def test_single_add(bx_dummy):
    from tests.real_responses.single_add import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.call(
        "tasks.task.add",
        {
            "fields": {
                "TITLE": "АГ",
                "DESCRIPTION": "Сложно сказать, почему потускнели светлые лики икон",
                "RESPONSIBLE_ID": 1,
                "TAGS": "скпэ22-4",
                "DEADLINE": "2022/07/30",
            }
        },
    )
    assert isinstance(results, dict)


def test_get_all_tasks(bx_dummy):
    from tests.real_responses.get_all_tasks import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_all(
        "tasks.task.list",
        {"filter": {"status": -1}, "select": ["ID", "STATUS", "STATUS_COMPLETE"]},
    )
    assert isinstance(results, list)
    assert len(results) == 1570


def test_crm_dealcategory_stage_list(bx_dummy):
    from tests.real_responses.crm_dealcategory_stage_list import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_by_ID("crm.dealcategory.stage.list", ["2", "4", "8", 0])
    assert isinstance(results, dict)


def test_crm_dealcategory_stage_list_batch_of_one(bx_dummy):
    from tests.real_responses.crm_dealcategory_stage_list_batch_of_one import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_by_ID("crm.dealcategory.stage.list", [0])

    # should return a list of dicts
    assert isinstance(results, list) and isinstance(results[0], dict)


def test_catalog_document_element_list(bx_dummy):
    from tests.real_responses.catalog_document_element_list import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_all("catalog.document.element.list")

    assert len(results) == 95


def test_crm_item_productrow_list(bx_dummy):
    from tests.real_responses.crm_item_productrow_list import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_all("crm.item.productrow.list")

    assert len(results) == 2


def test_crm_stagehistory_list(bx_dummy):
    from tests.real_responses.crm_stagehistory_list import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_all("crm.stagehistory.list")

    assert len(results) == 9


def test_socialnetwork_api_workgroup_list(bx_dummy):
    from tests.real_responses.socialnetwork_api_workgroup_list import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_all(
        "socialnetwork.api.workgroup.list",
        {
            "select": [
                "ID",
                "NAME",
                "DESCRIPTION",
                "DATE_CREATE",
                "DATE_UPDATE",
                "DATE_ACTIVITY",
                "SUBJECT_ID",
                "KEYWORDS",
                "IMAGE_ID",
                "NUMBER_OF_MEMBERS",
                "INITIATE_PERMS",
                "SPAM_PERMS",
                "SUBJECT_NAME",
            ]
        },
    )

    assert len(results) == 50


def test_crm_add_item(bx_dummy):
    from tests.real_responses.crm_item_add import response

    bx_dummy.srh = MockSRH(response)
    result = bx_dummy.call(
        "crm.item.add", {"entityTypeId": 1, "fields": {"first_name": "first_name"}}
    )
    assert result != "item" and isinstance(result, dict) and "id" in result


def test_crm_item_list(bx_dummy):
    from tests.real_responses.crm_item_list import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_all("crm.item.list")

    assert len(results) == 3



import pytest
from fast_bitrix24 import Bitrix
from fast_bitrix24.srh import ServerRequestHandler, TokenRejectedError

from typing import Dict, List, Union


class MockSRH(ServerRequestHandler):
    def __init__(self, token_func, response: Union[Dict, List[Dict]]):
        self.response = response if isinstance(response, list) else [response]
        self.element_no = 0

        super().__init__("https://google.com/path", token_func, False, 50, 2, 480, None)

    async def request_attempt(self, *args, **kwargs):
        result = self.response[self.element_no]
        try:
            if isinstance(result, Exception):
                raise result
            else:
                return result
        finally:
            self.element_no += 1


@pytest.mark.skip(reason="TODO")
def test_first_request():
    # нужно проверить, что вызывается функция запроса токена при первом запросе

    raise AssertionError


@pytest.mark.skip(reason="TODO")
def test_auth_success():
    # нужно проверить, что серверу передается токен, полученный от token_func

    raise AssertionError


@pytest.mark.skip(reason="TODO")
def test_auth_failure():
    # нужно проверить, что вызывается функция запроса токена, если сервер вернул ошибку токена

    raise AssertionError


@pytest.mark.skip(reason="TODO")
def test_abort_on_multiple_failures():
    # нужно проверить, что если token_func регулярно возвращает токен, который отвергается сервером,
    # то запрос оборвется после MAX_RETRIES неудачных попыток

    raise AssertionError


def test_expired_token(bx_dummy):
    # нужно проверить, что вызывается функция запроса токена, если токен истек

    called_count = 0

    async def token_func(*args, **kwargs):
        nonlocal called_count
        called_count += 1
        return "abc"

    bx_dummy.srh = MockSRH(token_func, [{}, TokenRejectedError(), {}])

    bx_dummy.call("test", raw=True)
    bx_dummy.call("test", raw=True)

    assert called_count == 2  # начальный запрос + запрос после истечения токена



import asyncio
import collections
import contextlib
import time

RequestRecord = collections.namedtuple("RequestRecord", "when, duration")


class SlidingWindowThrottler:
    """The class emulates a sliding window throttler.

    The consumer may only run requests until he has used up X seconds
    of request running time in total during a period of Y seconds.

    When the consumer has hit the limit, he will have to wait.
    """

    def __init__(self, max_request_running_time: float, measurement_period: float):
        # how much time fits into the bucket before it starts failing
        self._max_request_running_time = max_request_running_time

        # over what period of time should the max_request_running_time be measured
        self._measurement_period = measurement_period

        # request history. left - most recent, right - least recent
        self._request_history = collections.deque()

    @contextlib.asynccontextmanager
    async def acquire(self):
        """A context manager that will wait until it's safe to make the next request"""
        await asyncio.sleep(self._calculate_needed_sleep_time())

        try:
            yield
        finally:
            self._remove_stale_records()

    def _calculate_needed_sleep_time(self) -> float:
        """How much time to sleep before it's safe to make a request"""
        acc = 0
        for record in self._request_history:
            acc += record.duration
            if acc >= self._max_request_running_time:
                return record.when + self._measurement_period - time.monotonic()
        return 0

    def _remove_stale_records(self):
        """Remove all stale records from the record register"""
        cut_off = time.monotonic() - self._measurement_period
        while self._request_history and self._request_history[-1].when < cut_off:
            self._request_history.pop()

    def add_request_record(self, request_duration: float):
        """Register how long the last request has taken"""
        self._request_history.appendleft(
            RequestRecord(time.monotonic(), request_duration)
        )


class LeakyBucketThrottler:
    """The class implements a leaky bucket throttler.

    The consumer may only run requests until he has used up to X requests,
    after which the rate of Y requests per second will be applied.

    When the consumer has hit the limit, he will have to wait.
    """

    def __init__(self, pool_size: int, requests_per_second: float):
        # how many requests can be in the bucket at once
        self._pool_size = pool_size

        # how many requests are removed from the bucket per second
        self._requests_per_second = requests_per_second

        # request history. left - most recent, right - least recent
        self._request_history = collections.deque()

    @contextlib.asynccontextmanager
    async def acquire(self):
        """A context manager that will wait until it's safe to make the next request"""
        await asyncio.sleep(self._calculate_needed_sleep_time())

        try:
            yield
        finally:
            self._remove_stale_records()

    def _calculate_needed_sleep_time(self) -> float:
        """How much time to sleep before it's safe to make a request"""
        while len(self._request_history) >= self._pool_size:
            time_from_last_request = time.monotonic() - self._request_history[0]
            time_to_wait = 1 / self._requests_per_second - time_from_last_request
            if time_to_wait > 0:
                return time_to_wait
            else:
                break
        return 0

    def add_request_record(self):
        """Register when the last request was made"""
        self._request_history.appendleft(time.monotonic())

    def _remove_stale_records(self):
        """Remove all stale records from the record register"""
        cut_off = time.monotonic() - self._pool_size / self._requests_per_second
        while self._request_history and self._request_history[-1] < cut_off:
            self._request_history.pop()



from itertools import chain

from beartype.typing import Dict, List, Union


class ErrorInServerResponseException(Exception):
    pass


class ServerResponseParser:
    def __init__(self, response: dict, get_by_ID: bool = False):
        self.response = response
        self.get_by_ID = get_by_ID

    def more_results_expected(self) -> bool:
        return self.total and self.total > 50 and self.total != len(self.result)

    @property
    def result(self):
        return self.response.get("result")

    @property
    def total(self):
        return self.response.get("total")

    @property
    def error_description(self):
        return self.response.get("error_description")

    @property
    def result_error(self):
        return self.response.get("result_error")

    def extract_results(self) -> Union[Dict, List[Dict]]:
        """Вернуть результаты запроса.

        Если определено, что запрос был батчевым, то разобрать результаты батчей
        и собрать их в плоский список.

        Returns:
            Any: Результаты запроса, по возможности превращенные в плоский список.
        """
        self.raise_for_errors()

        if not self.is_batch():
            return self.extract_from_single_response(self.result)

        if not self.get_by_ID:
            return self.extract_from_batch_response(self.result["result"])

        extracted = self.extract_from_single_response(self.result["result"])

        return extracted[0] if isinstance(extracted, list) else extracted

    def raise_for_errors(self):
        errors = self.extract_errors()
        if errors:
            raise ErrorInServerResponseException(errors)

    def extract_errors(self):
        if self.is_batch():
            if self.result.get("result_error"):
                return self.result["result_error"]
        elif self.result_error:
            return self.result_error

        return None

    def is_batch(self) -> bool:
        return isinstance(self.result, dict) and "result" in self.result

    def extract_from_single_response(self, result: dict):
        # если результат вызова содержит только словарь из одного элемента {ключ: содержимое},
        # то вернуть это содержимое.
        # См. https://github.com/leshchenko1979/fast_bitrix24/issues/132

        # метод `crm.stagehistory.list` возвращает dict["items", list] --
        # разворачиваем его в список

        return next(iter(result.values())) if self.is_nested(result) else result

    @staticmethod
    def is_nested(result) -> bool:
        return isinstance(result, dict) and len(result) == 1

    def extract_from_batch_response(self, result) -> list:

        if not result:
            return []

        # если результат вызова содержит только словарь c одним ключом
        # и списком у него внутри, то вернуть этот список.
        # См. https://github.com/leshchenko1979/fast_bitrix24/issues/132

        first_item = next(iter(result.values()))

        # если внутри - списки, то вернуть их в одном плоском списке
        if self.is_nested(first_item) or isinstance(first_item, list):
            result_list = [
                self.extract_from_single_response(element)
                for element in result.values()
            ]

            # если получился вложенный список, то уплощаем
            if isinstance(result_list[0], list):
                result_list = list(chain(*result_list))

            return result_list

        # иначе (если внутри - dict), то вернуть в сам dict
        return result
